/* --------------------------------------------------------------- */
/* stmSysIO.c -- system-dependent core functions                   */
/* --------------------------------------------------------------- */
/*                                                                 */
/* Copyright (c) Mike Cowlishaw, 2010-2011.  All rights reserved.  */
/*                                                                 */
/* Permission to use, copy, modify, and/or distribute this         */
/* software for any purpose with or without fee is hereby granted, */
/* provided that the above copyright notice and this permission    */
/* notice appear in all copies, and that notice and the date of    */
/* any modifications be added to the software.                     */
/*                                                                 */
/* This software is provided "as is".  No warranties, whether      */
/* express, implied, or statutory, including, but not limited to,  */
/* implied warranties of merchantability and fitness for a         */
/* particular purpose apply to this software.  The author shall    */
/* not, in any circumstances, be liable for special, incidental,   */
/* or consequential damages, for any reason whatsoever.            */
/*                                                                 */
/* --------------------------------------------------------------- */
/* These are processor-specific functions for Tollos.              */
/*                                                                 */
/* They should not normally be used directly by applications.      */
/* --------------------------------------------------------------- */

#define  USEPLL 1             // set = 0 to use HSI clock for the
                              // system clock, e.g., for testing

#define  TOLLOS_IO            // allow use of I/O functions
#include "tollos.h"

#include "stm32f10x.h"
#include "stm32f10x_flash.h"
#include "stm32f10x_iwdg.h"
#include "stm32f10x_pwr.h"
#include "stm32f10x_rcc.h"

// Local statics
static int  msSysTick=0;      // SysTick interval; 0 until first set
static uint watchtime=0;      // Watchdog ms; 0 until first set

/* --------------------------------------------------------------- */
/* sysFeedDogIO -- feed the watchdog (if started)                  */
/*                                                                 */
/* Once the watchdog has been started, this function is called to  */
/* reset its counter and prevent a Watchdog Reset.  If called      */
/* before the watchdog is started it has no effect.                */
/* --------------------------------------------------------------- */
void sysFeedDogIO(void) {
  if (watchtime!=0) IWDG_ReloadCounter();
  } // sysFeedDogIO

/* --------------------------------------------------------------- */
/* sysInitProcIO -- initialize processor base function [startup]   */
/*                                                                 */
/*   returns 0 if OK, -1 if error (but probably cannot recover)    */
/*                                                                 */
/* This must be called before any other sysIO functions; it:       */
/*   -- initializes the processor [SystemInit or equivalent]       */
/*   -- sets the system core clock frequency (processor speed)     */
/*   -- maps the vector table to the start of RAM                  */
/*                                                                 */
/* Assumptions (cases catered for):                                */
/*   -- there is either an external crystal or oscillator, so      */
/*      &chipCrystal will be non-zero                              */
/*                                                                 */
/*   -- compile script defines one of the STM32 devices, e.g.,     */
/*      -DSTM32F10X_HD; this will need to change to make alternate */
/*      libraries for _VL and _CL lines, and the code below also   */
/*      may need modification for these                            */
/*                                                                 */
/*   -- stm32f10x.h has the configurable HSE_VALUE modification    */
/*      (see stmf10x.txt); this is required for RCC calculations   */
/*      for peripherals (and SystemCoreClockUpdate) -- however     */
/*      this does not fix SystemInit because this has hard-coded   */
/*      RCC_CFGR_PLLMULL9 multiplier                               */
/*                                                                 */
/*   -- speed will be set to highest of 48, 72, or 96 MHz if       */
/*      possible to allow setting of USB 48 MHz (STM32 devices do  */
/*      not currently support 96, but keeping the code the same as */
/*      for LPC)                                                   */
/*                                                                 */
/*   -- chipMaxClock will always be at least twice the external    */
/*      clock                                                      */
/* --------------------------------------------------------------- */
// Define required core clock speed
int sysInitProcIO(void) {     // processor initialize (clocks, etc.)
  uint sysclock;              // target clock speed

  /* ------------------------------------------------------------- */
  /* This is called after C initialization and the vector table    */
  /* has been copied to RAM.                                       */
  /* ------------------------------------------------------------- */
  RCC_DeInit();               // system reset (in case this is
                              // called other than after Reset)

  /* ------------------------------------------------------------- */
  /* Clocks setup ...                                              */
  /* ------------------------------------------------------------- */
  // SystemInit is not used because it does not allow oscillator
  // bypass (although in practice this is probably inconsequent) and
  // more seriously has a hard-coded RCC_CFGR_PLLMULL9 multiplier.
  //
  // At this point, the high speed internal oscillator (HSI) is
  // running; if USEPLL=0, system will use that, otherwise the PLL
  // frequency multiplier will be coaxed into action.

  // decide on the preferred target system clock speed
  if ((uint)&chipMaxClock>=96000000) sysclock=96000000;
   else if ((uint)&chipMaxClock>=72000000) sysclock=72000000;
   else if ((uint)&chipMaxClock>=48000000) sysclock=48000000;
   else sysclock=(uint)&chipMaxClock;

  #if USEPLL!=0
  { // declaration block
    ErrorStatus rs;                     // RCC return status
    uint extclock;                      // external clock speed
    uint pllmult;                       // PLL multiplier
    extclock=(uint)&chipCrystal;        // external clock speed
    if (!(uint)&chipOscillator) {       // do not have crystal
      RCC_HSEConfig(RCC_HSE_ON);        // start HSE, use chip oscillator
      }
     else {                             // have external oscillator
      RCC_HSEConfig(RCC_HSE_Bypass);    // start HSE, bypass chip oscillator
      }
    rs=RCC_WaitForHSEStartUp();         // wait until HSE is stable
    if (rs==ERROR) while(1);            // hang here for diagnosis

    // Enable flash prefetch + appropriate waitstates
    // (must be done before setting the clock)
    FLASH_PrefetchBufferCmd(FLASH_PrefetchBuffer_Enable);
    //   zero wait state, if 0 < SYSCLK <= 24 MHz
    //   one wait state,  if 24 MHz < SYSCLK <= 48 MHz
    //   two wait states, if 48 MHz < SYSCLK <= 72 MHz
    if (sysclock>48000000) FLASH_SetLatency(FLASH_Latency_2);
     else if (sysclock>24000000) FLASH_SetLatency(FLASH_Latency_1);
     else FLASH_SetLatency(FLASH_Latency_0);

    // set up derivative clocks
    RCC_HCLKConfig(RCC_SYSCLK_Div1);    // HCLK = Sys clock
    RCC_PCLK2Config(RCC_HCLK_Div1);     // PCLK2 = HCLK
    RCC_PCLK1Config(RCC_HCLK_Div2);     // PCLK1 = HCLK/2

    // set USB clock explicitly if running at 48 or 72 MHz
    if (sysclock==48000000) RCC_USBCLKConfig(RCC_USBCLKSource_PLLCLK_Div1);
     else if (sysclock==72000000) RCC_USBCLKConfig(RCC_USBCLKSource_PLLCLK_1Div5);

    // set up the PLL multiplier; this is the target speed divided by
    // the external clock speed, rounded down so as not to exceed the
    // maximum clock speed
    pllmult=sysclock/extclock;          // raw multiplier
    pllmult=(pllmult-2)<<18;            // massage for PLLConfig
    RCC_PLLConfig(RCC_PLLSource_HSE_Div1, pllmult);  // apply

    // start PLL and wait until ready
    RCC_PLLCmd(ENABLE);
    while(RCC_GetFlagStatus(RCC_FLAG_PLLRDY)==RESET);

    // set as system clock and wait until ready
    RCC_SYSCLKConfig(RCC_SYSCLKSource_PLLCLK);
    while (RCC_GetSYSCLKSource()!=RCC_CFGR_SWS_PLL);
    } // declaration
  #endif

  SystemCoreClockUpdate();              // Update CMSIS copy by
                                        // actual calculation

  /* ------------------------------------------------------------- */
  /* set vector table to RAM start and request 16 priority groups  */
  /* ------------------------------------------------------------- */
  SCB->VTOR=(uint32_t)&chipRAMbase & 0x3FFFFF80;
  NVIC_SetPriorityGrouping(3);          // all 16 levels preempt

  return 0;
  } // sysInitProcIO

/* --------------------------------------------------------------- */
/* sysIRQIO -- enable or disable an interrupt                      */
/*                                                                 */
/*   index  -- vector table index number (see sysVector; 0=WDT)    */
/*   enable -- 1 to enable, 0 to disable                           */
/*   level  -- priority level (if enabling) [0..15, 0=high]        */
/*                                                                 */
/*   returns 0 if OK, -1 if index or level is too large            */
/*                                                                 */
/* This function cannot be used to change the system vectors,      */
/* such as SysTick.                                                */
/* --------------------------------------------------------------- */
// This function should be identical for all CMSIS systems
int  sysIRQIO(uint index, flag enable, uint level) {

  if (index>=(uint)&chipVectors-16) return -1;

  if (enable) {
    if (level>SYSTEM_MINPRIO) return -1;
    NVIC_SetPriority(index, level);
    NVIC_EnableIRQ(index);
    }
   else {
    NVIC_DisableIRQ(index);
    }
  return 0;
  } // sysIRQIO

/* --------------------------------------------------------------- */
/* sysPowerIO -- set a power mode and wait for interrupt           */
/*                                                                 */
/*   item -- a constant listed in tollosIO.h                       */
/*                                                                 */
/* Valid items are:                                                */
/*   SYSTEM_DEEPPOWERDOWN -- power off until Reset or RTC wake     */
/*   SYSTEM_DEEPSLEEP     -- unclocked sleep                       */
/*   SYSTEM_POWERDOWN     -- power down (unclocked, slow start)    */
/*   SYSTEM_SLEEP         -- idle (Wait For Interrupt)             */
/*                                                                 */
/* Currently Tollos only uses SLEEP and DEEPPOWERDOWN.             */
/*                                                                 */
/* For STM32 DEEPSLEEP and POWERDOWN are the same (Stop mode).     */
/* --------------------------------------------------------------- */
void sysPowerIO(uint item)  {
  switch(item) {
    case SYSTEM_SLEEP: {           // Sleep (idle)
      __WFI();                     // wait for interrupt
      break;}
    case SYSTEM_DEEPSLEEP:         // Sleep (unclocked)
    case SYSTEM_POWERDOWN: {       // Power down (unclocked, slow start)
                                   // = STM32 Stop mode
      PWR_EnterSTOPMode(PWR_Regulator_LowPower, PWR_STOPEntry_WFI);
      break;}
    case SYSTEM_DEEPPOWERDOWN:{    // Power off until Reset or RTC
                                   // = STM32 Standby mode
      PWR_EnterSTANDBYMode();
      break;}
    default: {                     // ignore; return immediately
      break;}
    } // switch
  } // sysPowerIO


/* --------------------------------------------------------------- */
/* sysQueryIO -- query a system item                               */
/*                                                                 */
/*   item -- a constant listed in tollosIO.h                       */
/*                                                                 */
/*   returns current value of the item; negative generally         */
/*     indicates failure                                           */
/*                                                                 */
/* Valid items are:                                                */
/*   SYSTEM_CLOCK    -- clock speed, Hz                            */
/*   SYSTEM_RESET    -- 0 if last reset was normal (POR or Reset   */
/*                      pin), 1 if abnormal (Watchdog or Brownout) */
/*   SYSTEM_SYSTICK  -- SysTick interval in ms                     */
/*   SYSTEM_WATCHDOG -- Watchdog interval in ms                    */
/*                                                                 */
/* --------------------------------------------------------------- */
int sysQueryIO(uint item)  {
  int rc;                     // value to return
  switch(item) {

    case SYSTEM_CLOCK: {      // clock speed, Hz
      rc=SystemCoreClock;     // from the horse's mouth
      break;}

    case SYSTEM_RESET: {      // 0 = normal, 1 = abnormal
      // test for either Watchdog or brownout
      rc=(RCC->CSR &
          (RCC_CSR_IWDGRSTF | RCC_CSR_WWDGRSTF | RCC_CSR_LPWRRSTF)
         )!=0;
      break;}

    case SYSTEM_SYSTICK: {    // current SysTick interval
      rc=msSysTick;
      break;}

    case SYSTEM_WATCHDOG: {   // current Watchdog timeout
      rc=watchtime;
      break;}

    default: {
      rc=-1;                  // bad item
      break;}
    } // switch
  return rc;
  } // sysQueryIO

/* --------------------------------------------------------------- */
/* sysSetIO -- set a system item                                   */
/*                                                                 */
/*   item  -- a constant listed in tollosIO.h                      */
/*   value -- value to set the item to                             */
/*                                                                 */
/*   returns previous value of the item, or a negative number      */
/*     if an error or unknown                                      */
/*                                                                 */
/* Valid items are:                                                */
/*   SYSTEM_SYSTICK  -- SysTick interval in ms [>0, maximum        */
/*                      depends on clock rate]                     */
/*   SYSTEM_WATCHDOG -- Watchdog interval in ms                    */
/*                                                                 */
/* --------------------------------------------------------------- */
int sysSetIO(uint item, int value)  {
  int rc;                // value to return
  switch(item) {

    case SYSTEM_SYSTICK: {              // set SysTick [ms]
      uint count;                       // work
      rc=msSysTick;

      // check parameter
      if (value<=0) return -1;          // must be at least 1

      // calculate SysTick counter (number to divide clock by); this
      // must be no more than 0xFFFFFF; calculate carefully to
      // avoid multiplication overflow
      count=SystemCoreClock/1000 * value;
      if (count>0xFFFFFF) return -1;    // SysTick counter out of range
      // count=72000;  // (this would force diviisor for 72 MHz sys clock)

      // The CMSSIS quick setter (also sets to lowest priority)
      if (SysTick_Config(count)) rc=-1; // failed
       else msSysTick=value;            // good -- record it
      break;}

    case SYSTEM_WATCHDOG: {             // watchdog timeout
      uint ms=(uint)value;
      if (ms>TOLLOS_MAXWATCHDOG || ms<1) return -1;  // out of range
      rc=watchtime;

      // STM32 IWDG runs from a clock which can vary from 30-60kHz,
      // so the time calculation here uses the worst case (fastest
      // clock) to make sure that the watchdog period is at least as
      // long as requested; for simplicity, 64kHz is assumed

      // enable write access to IWDG_PR and IWDG_RLR registers
      IWDG->KR=IWDG_WriteAccess_Enable;
      // prescale /246 so that counter runs at 250Hz (4ms/count)
      // giving a maximum time of ~16 secs
      IWDG->PR=IWDG_Prescaler_256;
      IWDG->RLR=(uint)value/4+1;        // set reload from ms
      // no equates for KR_KEY_Reload or _Enable, so must call
      // functions, here
      IWDG_ReloadCounter();             // .. and reload
      IWDG_Enable();

      watchtime=value;
      break;}

    default: {
      rc=-1;             // bad item
      break;}
    } // switch
  return rc;
  } // sysSetIO

